from dune.perftool.pdelab.argument import (name_accumulation_variable,
                                           PDELabAccumulationFunction,
                                           )
from dune.perftool.generation import (backend,
                                      domain,
                                      dump_accumulate_timer,
                                      function_mangler,
                                      get_counted_variable,
                                      get_counter,
                                      iname,
                                      instruction,
                                      post_include,
                                      kernel_cached,
                                      temporary_variable,
                                      transform,
                                      )
from dune.perftool.options import get_option
from dune.perftool.loopy.flatten import flatten_index
from dune.perftool.loopy.buffer import get_buffer_temporary
from dune.perftool.sumfact.quadrature import nest_quadrature_loops
from dune.perftool.pdelab.localoperator import determine_accumulation_space
from dune.perftool.pdelab.restriction import restricted_name
from dune.perftool.pdelab.signatures import assembler_routine_name
from dune.perftool.pdelab.geometry import world_dimension
from dune.perftool.sumfact.tabulation import (basis_functions_per_direction,
                                              construct_basis_matrix_sequence,
                                              )
from dune.perftool.sumfact.switch import (get_facedir,
                                          get_facemod,
                                          )
from dune.perftool.sumfact.symbolic import SumfactKernel
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.tools import get_pymbolic_basename
from dune.perftool.error import PerftoolError
from dune.perftool.sumfact.quadrature import quadrature_inames


import loopy as lp
import numpy as np
import pymbolic.primitives as prim


@iname
def _sumfact_iname(bound, _type, count):
    name = "sf_{}_{}".format(_type, str(count))
    domain(name, bound)
    return name


def sumfact_iname(bound, _type):
    count = get_counter('_sumfac_iname_{}'.format(_type))
    name = _sumfact_iname(bound, _type, count)
    return name


@kernel_cached
def accum_iname(restriction, bound, i):
    return sumfact_iname(bound, "accum")


def name_test_function_contribution(test):
    count = get_counter('sumfact_contribution_counter')
    grad = "grad_" if test.reference_grad else ""
    return restricted_name("contrib_{}phi_{}".format(grad, str(count)), test.restriction)


@backend(interface="accum_insn", name="sumfact")
def generate_accumulation_instruction(visitor, accterm, measure, subdomain_id):
    pymbolic_expr = visitor(accterm.term)

    if pymbolic_expr == 0:
        return

    dim = world_dimension()

    # If this is a gradient, we find the gradient iname
    additional_inames = frozenset({})
    if accterm.new_indices is not None:
        for i in accterm.new_indices:
            if i not in visitor.dimension_indices:
                from dune.perftool.pdelab.localoperator import grad_iname
                additional_inames = additional_inames.union(frozenset({grad_iname(i, dim)}))

    # Figure out the name of the accumulation variable
    test_lfs = determine_accumulation_space(accterm.argument.expr, 0, measure)
    ansatz_lfs = determine_accumulation_space(accterm.term, 1, measure)
    accum = name_accumulation_variable(test_lfs.get_restriction() + ansatz_lfs.get_restriction())

    def emit_sumfact_kernel(i, restriction, insn_dep):
        # Construct the matrix sequence for this sum factorization
        matrix_sequence = construct_basis_matrix_sequence(transpose=True,
                                                          derivative=i if accterm.new_indices else None,
                                                          facedir=get_facedir(accterm.argument.restriction),
                                                          facemod=get_facemod(accterm.argument.restriction),
                                                          )

        sf = SumfactKernel(matrix_sequence=matrix_sequence,
                           restriction=(accterm.argument.restriction, restriction),
                           stage=3,
                           preferred_position=i if accterm.new_indices else None,
                           within_inames=visitor.inames,
                           accumvar=accum,
                           )

        from dune.perftool.sumfact.vectorization import attach_vectorization_info
        vsf = attach_vectorization_info(sf)

        # Make sure we have a buffer that we can set up the input with
        buffer = vsf.buffer
        if buffer is None:
            buffer = get_counted_variable("buffer")

        vectag = frozenset({"gradvec"}) if vsf.vectorized else frozenset()

        temp = get_buffer_temporary(buffer,
                                    shape=vsf.quadrature_shape,
                                    dim_tags=vsf.quadrature_dimtags,
                                    name=vsf.input,
                                    )

        # Those input fields, that are padded need to be set to zero
        # in order to do a horizontal_add later on
        for pad in vsf.padding:
            assignee = prim.Subscript(prim.Variable(temp),
                                      tuple(prim.Variable(i) for i in quadrature_inames()) + (pad,))
            instruction(assignee=assignee,
                        expression=0,
                        forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                        forced_iname_deps_is_final=True,
                        tags=frozenset(["quadvec", "gradvec"])
                        )

        # Replace gradient iname with correct index for assignement
        replace_dict = {}
        for iname in additional_inames:
            replace_dict[prim.Variable(iname)] = i
        from dune.perftool.loopy.symbolic import substitute
        expression = substitute(pymbolic_expr, replace_dict)

        # Write timing stuff for jacobian (for alpha methods it is done at the end of stage 1)
        timer_dep = frozenset()
        if get_option("instrumentation_level") >= 4:
            timer_name = assembler_routine_name() + '_kernel' + '_quadratureloop'
            post_include('HP_DECLARE_TIMER({});'.format(timer_name), filetag='operatorfile')
            dump_accumulate_timer(timer_name)
            if(visitor.inames):
                timer_dep = frozenset({instruction(code="HP_TIMER_START({});".format(timer_name),
                                                   within_inames=frozenset(visitor.inames))})

        # Determine dependencies
        from loopy.match import Or, Writes
        from loopy.symbolic import DependencyMapper
        from dune.perftool.tools import get_pymbolic_basename
        deps = Or(tuple(Writes(get_pymbolic_basename(expr)) for expr in DependencyMapper()(expression)))

        # Issue an instruction in the quadrature loop that fills the buffer
        # with the evaluation of the contribution at all quadrature points
        assignee = prim.Subscript(prim.Variable(temp),
                                  tuple(prim.Variable(i) for i in quadrature_inames()) + vsf.vec_index_tuple(sf))
        contrib_dep = instruction(assignee=assignee,
                                  expression=expression,
                                  forced_iname_deps=frozenset(quadrature_inames() + visitor.inames),
                                  forced_iname_deps_is_final=True,
                                  tags=frozenset({"quadvec"}).union(vectag),
                                  depends_on=frozenset({deps}).union(timer_dep)
                                  )

        if insn_dep is None:
            insn_dep = frozenset({contrib_dep})

        if get_option("instrumentation_level") >= 4:
            insn_dep = frozenset({instruction(code="HP_TIMER_STOP({});".format(timer_name),
                                              depends_on=insn_dep,
                                              within_inames=frozenset(visitor.inames))})

        inames = tuple(accum_iname((accterm.argument.restriction, restriction), mat.rows, i)
                       for i, mat in enumerate(vsf.matrix_sequence))

        # Collect the lfs and lfs indices for the accumulate call
        test_lfs.index = flatten_index(tuple(prim.Variable(i) for i in inames),
                                       (basis_functions_per_direction(),) * dim,
                                       order="f"
                                       )

        # In the jacobian case, also determine the space for the ansatz space
        rank = 2 if visitor.inames else 1
        if rank == 2:
            # TODO the next line should get its inames from
            # elsewhere. This is *NOT* robust (but works right now)
            ansatz_lfs.index = flatten_index(tuple(prim.Variable(visitor.inames[i])
                                                   for i in range(world_dimension())),
                                             (basis_functions_per_direction(),) * dim,
                                             order="f"
                                             )

        # Add a sum factorization kernel that implements the multiplication
        # with the test function (stage 3)
        from dune.perftool.sumfact.realization import realize_sum_factorization_kernel
        result, insn_dep = realize_sum_factorization_kernel(vsf.copy(insn_dep=insn_dep))

        # Determine the expression to accumulate with. This depends on the vectorization strategy!
        result = prim.Subscript(result, tuple(prim.Variable(i) for i in inames))
        vecinames = ()
        # TODO: evaluate whether the following line would be okay with vsf.vectorized
        if vsf.vec_index(sf) is not None:
            iname = accum_iname((accterm.argument.restriction, restriction), 4, "vec")
            vecinames = (iname,)
            transform(lp.tag_inames, [(iname, "vec")])
            from dune.perftool.tools import maybe_wrap_subscript
            result = prim.Call(prim.Variable("horizontal_add"),
                               (maybe_wrap_subscript(result, prim.Variable(iname)),),
                               )

        if not get_option("fastdg"):
            expr = prim.Call(PDELabAccumulationFunction(accum, rank),
                             (test_lfs.get_args() +
                              ansatz_lfs.get_args() +
                              (result,)
                              )
                             )
            instruction(assignees=(),
                        expression=expr,
                        forced_iname_deps=frozenset(inames + visitor.inames + vecinames),
                        forced_iname_deps_is_final=True,
                        depends_on=insn_dep,
                        )

        # Mark the transformation that moves the quadrature loop
        # inside the trialfunction loops for application
        transform(nest_quadrature_loops, visitor.inames)

        return insn_dep

    # Extract the restrictions on argument-1:
    jac_restrictions = frozenset(tuple(ma.restriction for ma in
                                       extract_modified_arguments(accterm.term, argnumber=1, do_index=True)))
    if not jac_restrictions:
        jac_restrictions = frozenset({0})

    insn_dep = None
    for restriction in jac_restrictions:
        if accterm.new_indices:
            for i in range(world_dimension()):
                insn_dep = emit_sumfact_kernel(i, restriction, insn_dep)
        else:
            insn_dep = emit_sumfact_kernel(None, restriction, insn_dep)
