""" A pymbolic node representing a sum factorization kernel """

from dune.perftool.options import get_option
from dune.perftool.generation import get_counted_variable
from dune.perftool.sumfact.tabulation import BasisTabulationMatrixArray

from pytools import ImmutableRecord, product

import pymbolic.primitives as prim
import loopy as lp
import frozendict
import inspect


class SumfactKernelBase(object):
    pass


class SumfactKernel(SumfactKernelBase, ImmutableRecord, prim.Variable):
    def __init__(self,
                 matrix_sequence=None,
                 buffer=None,
                 stage=1,
                 preferred_position=None,
                 restriction=0,
                 within_inames=(),
                 input=None,
                 insn_dep=frozenset(),
                 coeff_func=None,
                 element=None,
                 component=None,
                 accumvar=None,
                 ):
        """Create a sum factorization kernel

        Sum factorization can be written as

        Y = R_{d-1} (A_{d-1} * ... * R_0 (A_0 * X)...)

        with:
        - X: Input rank d tensor of dimension n_0 x ... x n_{d-1}
        - Y: Output rank d tensor of dimension m_0 x ... x m_{d-1}
        - A_l: Values of 1D basis evaluations at quadrature points in l
               direction, matrix of dimension m_l x n_l
        - R_l: Transformation operator that permutes the underlying data
               vector of the rank d tensor in such a way that the fastest
               direction gets the slowest direction

        In the l'th step we have the following setup:
        - A_l: Matrix of dimensions m_l x n_l
        - X_l: Rank d tensor of dimensions n_l x ... x n_{d-1} x m_0 x ... x m_{l-1}
        - R_l: Transformation operator

        Looking at the indizes the following will happen:
        X --> [n_l,...,n_{d-1},m_0,...,m_{l-1}]
        A_l * X --> [m_l,n_l] * [n_l, ...] = [m_l,n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]
        R_l (A_l*X) --> [n_{l+1},...,n_{d-1},m_0,...,m_{l-1}]

        So the multiplication with A_l is a reduction over one index and
        the transformation brings the next reduction index in the fastest
        position.

        It can make sense to permute the order of directions. If you have
        a small m_l (e.g. stage 1 on faces) it is better to do direction l
        first. This can be done by:

        - Permuting the order of the A matrices.
        - Permuting the input tensor.
        - Permuting the output tensor (this assures that the directions of
          the output tensor are again ordered from 0 to d-1).

        Note, that you will typically *not* set all of the below arguments,
        but only some. The vectorization strategy may set others for you.
        The only argument really needed in all cases is matrix_sequence.

        Arguments:
        ----------
        matrix_sequence: A tuple of BasisTabulationMatrixBase instances
            The list of tensors to be applied to the input.
            Order of application is from 0 up.
        buffer: A string identifying the flip flop buffer in use
            for intermediate results. The memory is expected to be
            pre-initialized with the input or you have to provide
            direct_input (FastDGGridOperator).
        stage: 1 or 3
        insn_dep: An instruction ID that the first issued instruction
            should depend upon. All following ones will depend on each
            other.
        within_inames: Instructions will be executed within those
            inames (needed for stage 3 in jacobians).
        preferred_position: Will be used in the dry run to order kernels
            when doing vectorization e.g. (dx u,dy u,dz u, u).
        restriction: Restriction for faces values.
        input: The name to use for the input temporary
        padding: a set of slots in the input temporary to be padded
        index: The slot in the input temporary dedicated to this kernel
        coeff_func: The function to call to get the input coefficient
        element: The UFL element
        component: The treepath to the correct component of above element
        accumvar: The accumulation variable to accumulate into
        """
        # Assert the inputs!
        assert isinstance(matrix_sequence, tuple)
        from dune.perftool.sumfact.tabulation import BasisTabulationMatrixBase
        assert all(isinstance(m, BasisTabulationMatrixBase) for m in matrix_sequence)

        assert stage in (1, 3)

        if preferred_position is not None:
            assert isinstance(preferred_position, int)

        if stage == 3:
            assert isinstance(restriction, tuple)

        assert isinstance(within_inames, tuple)

        assert isinstance(insn_dep, frozenset)

        # The following construction is a bit weird: Dict comprehensions do not have
        # access to the locals of the calling scope: So we need to do the eval beforehand
        defaults = [eval(a) for a in SumfactKernel.init_arg_names]
        defaultdict = {a: defaults[SumfactKernel.init_arg_names.index(a)] for a in SumfactKernel.init_arg_names}

        # Call the base class constructors
        ImmutableRecord.__init__(self, **defaultdict)
        prim.Variable.__init__(self, "SUMFACT")

    #
    # The methods/fields needed to get a well-formed pymbolic node
    #

    def __getinitargs__(self):
        return tuple(getattr(self, arg) for arg in SumfactKernel.init_arg_names)

    def stringifier(self):
        return lp.symbolic.StringifyMapper

    mapper_method = "map_sumfact_kernel"

    #
    # Some cache key definitions
    # Watch out for the documentation to see which key is used unter what circumstances
    #

    @property
    def cache_key(self):
        """ The cache key that can be used in generation magic
        Any two sum factorization kernels having the same cache_key
        are realized simulatenously!
        """
        return (self.matrix_sequence, self.restriction, self.stage, self.buffer)

    @property
    def input_key(self):
        """ A cache key for the input coefficients
        Any two sum factorization kernels having the same input_key
        work on the same input coefficient (and are suitable for simultaneous
        treatment because of that)
        """
        return (self.restriction, self.stage, self.coeff_func, self.element, self.component, self.accumvar)

    #
    # Some convenience methods to extract information about the sum factorization kernel
    #

    @property
    def length(self):
        """ The number of matrices to apply """
        return len(self.matrix_sequence)

    @property
    def vectorized(self):
        return False

    @property
    def transposed(self):
        return next(iter(self.matrix_sequence)).transpose

    def vec_index(self, sf):
        """ Map an unvectorized sumfact kernel object to its position
        in the vectorized kernel
        """
        return None

    def vec_index_tuple(self, sf):
        return ()

    @property
    def flat_input_shape(self):
        """ The 'flat' input tensor shape """
        return (product(mat.cols for mat in self.matrix_sequence),)

    @property
    def quadrature_shape(self):
        """ The shape of a temporary for the quadrature points

        Takes into account the lower dimensionality of faces and vectorization.
        """
        if self.transposed:
            return tuple(mat.cols for mat in self.matrix_sequence if mat.face is None)
        else:
            return tuple(mat.rows for mat in self.matrix_sequence if mat.face is None)

    @property
    def quadrature_dimtags(self):
        """ The dim_tags of a temporary for the quadrature points

        Takes into account the lower dimensionality of faces and vectorization.
        """
        tags = ["f"] * len(self.quadrature_shape)
        return ",".join(tags)

    @property
    def dof_shape(self):
        """ The shape of a temporary for the degrees of freedom

        Takes into account vectorization.
        """
        shape = tuple(mat.rows for mat in self.matrix_sequence)
        return shape

    @property
    def dof_dimtags(self):
        """ The dim_tags of a temporary for the degrees of freedom

        Takes into account vectorization.
        """
        tags = ["f"] * len(self.dof_shape)
        return ",".join(tags)

    @property
    def output_shape(self):
        if self.stage == 1:
            return self.quadrature_shape
        else:
            return self.dof_shape

    @property
    def output_dimtags(self):
        if self.stage == 1:
            return self.quadrature_dimtags
        else:
            return self.dof_dimtags

    def output_to_pymbolic(self, name):
        return lp.TaggedVariable(name, "sumfac")

    #
    # Define properties for conformity with the interface of VectorizedSumfactKernel
    #

    @property
    def padding(self):
        return set()


# Extract the argument list and store it on the class. This needs to be done
# outside of the class because the SumfactKernel class object needs to be fully
# initialized in order to extract the information from __init__.
SumfactKernel.init_arg_names = tuple(inspect.getargspec(SumfactKernel.__init__)[0][1:])


class VectorizedSumfactKernel(SumfactKernelBase, ImmutableRecord, prim.Variable):
    def __init__(self,
                 kernels=None,
                 vector_width=get_option('max_vector_width'),
                 buffer=None,
                 input=None,
                 insn_dep=frozenset(),
                 ):
        # Assert the input data structure
        assert isinstance(kernels, tuple)
        assert all(isinstance(k, SumfactKernel) for k in kernels)

        # Assert all the properties that need to be the same across all subkernels
        assert len(set(k.stage for k in kernels)) == 1
        assert len(set(k.length for k in kernels)) == 1
        assert len(set(k.restriction for k in kernels)) == 1
        assert len(set(k.within_inames for k in kernels)) == 1

        # Assert properties of the matrix sequence of the underlying kernels
        for i in range(kernels[0].length):
            assert len(set(tuple(k.matrix_sequence[i].rows for k in kernels))) == 1
            assert len(set(tuple(k.matrix_sequence[i].cols for k in kernels))) == 1
            assert len(set(tuple(k.matrix_sequence[i].face for k in kernels))) == 1
            assert len(set(tuple(k.matrix_sequence[i].transpose for k in kernels))) == 1

        # Join the instruction dependencies of all subkernels
        insn_dep = insn_dep.union(k.insn_dep for k in kernels)

        # We currently assume that all subkernels are consecutive, 0-based within the vector
        assert None not in kernels

        ImmutableRecord.__init__(self,
                                 kernels=kernels,
                                 vector_width=vector_width,
                                 buffer=buffer,
                                 input=input,
                                 insn_dep=insn_dep,
                                 )

        prim.Variable.__init__(self, "VecSUMFAC")

    def __getinitargs__(self):
        return (self.kernels, self.vector_width, self.buffer, self.input, self.insn_dep)

    def stringifier(self):
        return lp.symbolic.StringifyMapper

    mapper_method = "map_vectorized_sumfact_kernel"

    init_arg_names = ("kernels", "vector_width", "buffer", "input", "insn_dep")

    #
    # Some cache key definitions
    # Watch out for the documentation to see which key is used unter what circumstances
    #

    @property
    def cache_key(self):
        """ The cache key that can be used in generation magic
        Any two sum factorization kernels having the same cache_key
        are realized simulatenously!
        """
        return (self.matrix_sequence, self.restriction, self.stage, self.buffer)

    @property
    def input_key(self):
        """ A cache key for the input coefficients
        Any two sum factorization kernels having the same input_key
        work on the same input coefficient (and are suitable for simultaneous
        treatment because of that)
        """
        return (self.restriction, self.stage, self.coeff_func, self.element, self.component, self.accumvar)

    #
    # Deduce all data fields of normal sum factorization kernels from the underlying kernels
    #

    @property
    def matrix_sequence(self):
        return tuple(BasisTabulationMatrixArray(rows=self.kernels[0].matrix_sequence[i].rows,
                                                cols=self.kernels[0].matrix_sequence[i].cols,
                                                transpose=self.kernels[0].matrix_sequence[i].transpose,
                                                face=self.kernels[0].matrix_sequence[i].face,
                                                derivative=tuple(k.matrix_sequence[i].derivative for k in self.kernels),
                                                )
                     for i in range(self.length))

    @property
    def stage(self):
        return self.kernels[0].stage

    @property
    def restriction(self):
        return self.kernels[0].restriction

    @property
    def within_inames(self):
        return self.kernels[0].within_inames

    @property
    def coeff_func(self):
        assert len(set(k.coeff_func for k in self.kernels)) == 1
        return self.kernels[0].coeff_func

    @property
    def element(self):
        assert len(set(k.element for k in self.kernels)) == 1
        return self.kernels[0].element

    @property
    def component(self):
        assert len(set(k.component for k in self.kernels)) == 1
        return self.kernels[0].component

    @property
    def accumvar(self):
        assert len(set(k.accumvar for k in self.kernels)) == 1
        return self.kernels[0].accumvar

    #
    # Define some properties only needed for this one
    #

    @property
    def padding(self):
        return set(range(self.vector_width)) - set(range(len(self.kernels)))

    #
    # Define the same properties the normal SumfactKernel defines
    #

    @property
    def cache_key(self):
        return tuple(k.cache_key for k in self.kernels)

    @property
    def input_key(self):
        return tuple(k.input_key for k in self.kernels)

    @property
    def length(self):
        return self.kernels[0].length

    @property
    def vectorized(self):
        return True

    @property
    def transposed(self):
        return self.kernels[0].transposed

    def vec_index(self, sf):
        return self.kernels.index(sf)

    def vec_index_tuple(self, sf):
        return (self.vec_index(sf),)

    @property
    def flat_input_shape(self):
        return (product(mat.cols for mat in self.matrix_sequence), 4)

    @property
    def quadrature_shape(self):
        if self.transposed:
            return tuple(mat.cols for mat in self.matrix_sequence if mat.face is None) + (4,)
        else:
            return tuple(mat.rows for mat in self.matrix_sequence if mat.face is None) + (4,)

    @property
    def quadrature_dimtags(self):
        tags = ["f"] * len(self.quadrature_shape)
        tags[-1] = 'c'
        return ",".join(tags)

    @property
    def dof_shape(self):
        return tuple(mat.rows for mat in self.matrix_sequence) + (4,)

    @property
    def dof_dimtags(self):
        tags = ["f"] * len(self.dof_shape)
        tags[-1] = 'vec'
        return ",".join(tags)

    @property
    def output_shape(self):
        if self.stage == 1:
            return self.quadrature_shape
        else:
            return self.dof_shape

    @property
    def output_dimtags(self):
        if self.stage == 1:
            return self.quadrature_dimtags
        else:
            return self.dof_dimtags

    def output_to_pymbolic(self, name):
        return lp.TaggedVariable(name, "vector")
